///////////////////////////////////////////////////////////////////////////////
// main.cpp
// ========
// spiral pipe
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2016-04-12
// UPDATED: 2022-12-24
///////////////////////////////////////////////////////////////////////////////

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cstdlib>
#include <cstring>                      // std::memcpy()
#include <iostream>
#include <sstream>
#include <string>
#include <iomanip>
#include <vector>
#include "glExtension.h"                // helper for OpenGL extensions
#include "Timer.h"
#include "Vectors.h"
#include "Matrices.h"
#include "Plane.h"
#include "Line.h"
#include "Pipe.h"



// GLUT CALLBACK functions ////////////////////////////////////////////////////
void displayCB();
void reshapeCB(int w, int h);
void timerCB(int millisec);
void idleCB();
void keyboardCB(unsigned char key, int x, int y);
void mouseCB(int button, int stat, int x, int y);
void mouseMotionCB(int x, int y);
void mousePassiveMotionCB(int x, int y);

// CALLBACK function when exit() called ///////////////////////////////////////
void exitCB();

// function declearations /////////////////////////////////////////////////////
void initGL();
bool initGLSL();
int  initGLUT(int argc, char **argv);
bool initSharedMem();
void clearSharedMem();
void initLights();
void setCamera(float posX, float posY, float posZ, float targetX, float targetY, float targetZ);
void drawString(const char *str, int x, int y, float color[4], void *font);
void drawString3D(const char *str, float pos[3], float color[4], void *font);
void showInfo();
void toOrtho();
void toPerspective();
void drawPipe();
void drawPath();
void draw();
std::vector<Vector3> buildSpiralPath(float r1, float r2, float h1, float h2, float turns, int points);
std::vector<Vector3> buildCircle(float radius, int steps);
void buildVbo();


// constants
const int SCREEN_WIDTH = 600;
const int SCREEN_HEIGHT = 600;
const int CIRCLE_SECTORS = 100;     // # of vertices per contour
const int PATH_POINTS = 300;        // # of points on the path
const float PATH_TURNS = 4.5f;      // # of turns of the spiral


// flat shading ===========================================
const char* vsSource1 = R"(
// GLSL version
#version 110
// uniforms
uniform mat4 matrixModelViewProjection;
// vertex attribs (inputs)
attribute vec3 vertexPosition;
void main()
{
    gl_Position = matrixModelViewProjection * vec4(vertexPosition, 1.0);
}
)";
const char* fsSource1 = R"(
// GLSL version
#version 110
// uniforms
uniform vec4 materialDiffuse;
void main()
{
    gl_FragColor = materialDiffuse;
}
)";


// blinn shading with texture =============================
const char* vsSource2 = R"(
// GLSL version
#version 110
// uniforms
uniform mat4 matrixModelView;
uniform mat4 matrixNormal;
uniform mat4 matrixModelViewProjection;
// vertex attribs (input)
attribute vec3 vertexPosition;
attribute vec3 vertexNormal;
attribute vec2 vertexTexCoord;
// varyings (output)
varying vec3 esVertex, esNormal;
varying vec2 texCoord0;
void main()
{
    esVertex = vec3(matrixModelView * vec4(vertexPosition, 1.0));
    esNormal = vec3(matrixNormal * vec4(vertexNormal, 1.0));
    texCoord0 = vertexTexCoord;
    gl_Position = matrixModelViewProjection * vec4(vertexPosition, 1.0);
}
)";

const char* fsSource2 = R"(
// GLSL version
#version 110
// uniforms
uniform vec4 lightPosition;             // should be in the eye space
uniform vec4 lightAmbient;              // light ambient color
uniform vec4 lightDiffuse;              // light diffuse color
uniform vec4 lightSpecular;             // light specular color
uniform vec4 materialAmbient;           // material ambient color
uniform vec4 materialDiffuse;           // material diffuse color
uniform vec4 materialSpecular;          // material specular color
uniform float materialShininess;        // material specular shininess
uniform sampler2D map0;                 // texture map #1
uniform bool textureUsed;               // flag for texture
// varyings
varying vec3 esVertex, esNormal;
varying vec2 texCoord0;
void main()
{
    vec3 normal = normalize(esNormal);
    vec3 light;
    if(lightPosition.w == 0.0)
    {
        light = normalize(lightPosition.xyz);
    }
    else
    {
        light = normalize(lightPosition.xyz - esVertex);
    }
    vec3 view = normalize(-esVertex);
    vec3 halfv = normalize(light + view);

    vec3 color = lightAmbient.rgb * materialAmbient.rgb;        // begin with ambient
    float dotNL = max(dot(normal, light), 0.0);
    color += lightDiffuse.rgb * materialDiffuse.rgb * dotNL;    // add diffuse
    if(textureUsed)
        color *= texture2D(map0, texCoord0).rgb;                // modulate texture map
    float dotNH = max(dot(normal, halfv), 0.0);
    color += pow(dotNH, materialShininess) * lightSpecular.rgb * materialSpecular.rgb; // add specular

    // set frag color
    gl_FragColor = vec4(color, materialDiffuse.a);
}
)";



// global variables
void *font = GLUT_BITMAP_8_BY_13;
bool mouseLeftDown;
bool mouseRightDown;
float mouseX, mouseY;
float cameraAngleX;
float cameraAngleY;
float cameraDistance = 15;
int screenWidth, screenHeight;
int drawMode;
int currIndex = 0;
bool animating = true;
Matrix4 matrixModelView;
Matrix4 matrixProjection;
Timer t1, t2;               // to measure copying data to VBO
std::vector<Vector3> path;
std::vector<Vector3> circle;
Pipe pipe;

// for VBO
bool vboSupported;
GLuint vboId = 0;       // ID of VBO for vertex arrays
GLuint iboId = 0;       // ID of VBO for index array

// for GLSL
bool glslSupported;
bool uniformLocationSupported;
GLuint progId1 = 0;     // IDs of GLSL programs
GLuint progId2 = 0;
// uniforms
GLint uniformMatrixModelView;
GLint uniformMatrixModelViewProjection;
GLint uniformMatrixNormal;
GLint uniformLightPosition;
GLint uniformLightAmbient;
GLint uniformLightDiffuse;
GLint uniformLightSpecular;
GLint uniformMaterialAmbient;
GLint uniformMaterialDiffuse;
GLint uniformMaterialSpecular;
GLint uniformMaterialShininess;
GLint uniformMap0;
GLint uniformTextureUsed;
// attributes
GLint attribVertexPosition;
GLint attribVertexNormal;
GLint attribVertexTexCoord;



///////////////////////////////////////////////////////////////////////////////
// draw a pipe
///////////////////////////////////////////////////////////////////////////////
void drawPipe()
{
    if(!glslSupported)
        return;

    // bind GLSL
    glUseProgram(progId2);
    glBindTexture(GL_TEXTURE_2D, 0);

    // activate attribs
    glEnableVertexAttribArray(attribVertexPosition);
    glEnableVertexAttribArray(attribVertexNormal);

    // set attrib arrays using glVertexAttribPointer()
    int offset = sizeof(float) * 3 * (int)pipe.getContour(0).size() * pipe.getContourCount();
    glBindBuffer(GL_ARRAY_BUFFER, vboId);
    glVertexAttribPointer(attribVertexPosition, 3, GL_FLOAT, false, 0, 0);
    glVertexAttribPointer(attribVertexNormal, 3, GL_FLOAT, false, 0, (void*)offset);

    // transform camera (view)
    Matrix4 matrixView;
    matrixView.rotateY(cameraAngleY);
    matrixView.rotateX(cameraAngleX);
    matrixView.translate(0, 0, -cameraDistance);

    // set model matrix uniforms
    matrixModelView = matrixView;
    Matrix4 matrixModelViewProjection = matrixProjection * matrixModelView;
    Matrix4 matrixNormal = matrixModelView;
    matrixNormal.setColumn(3, Vector4(0,0,0,1));

    // set uniforms
    glUniformMatrix4fv(uniformMatrixModelView, 1, false, matrixModelView.get());
    glUniformMatrix4fv(uniformMatrixModelViewProjection, 1, false, matrixModelViewProjection.get());
    glUniformMatrix4fv(uniformMatrixNormal, 1, false, matrixNormal.get());
    glUniform1i(uniformMap0, 0);
    glUniform1i(uniformTextureUsed, 0);

    // draw with triangle strip
    int indexSize = 2 * (pipe.getContourCount()-1) * pipe.getContour(0).size();
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboId);
    glDrawElements(GL_TRIANGLE_STRIP, indexSize, GL_UNSIGNED_INT, 0);

    glDisableVertexAttribArray(attribVertexPosition);
    glDisableVertexAttribArray(attribVertexNormal);

    // unbind
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glUseProgram(0);
}



///////////////////////////////////////////////////////////////////////////////
// draw lines along the path
///////////////////////////////////////////////////////////////////////////////
void drawPath()
{
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);

    glPushMatrix();
    glTranslatef(0, 0, -cameraDistance);
    glRotatef(cameraAngleX, 1, 0, 0);
    glRotatef(cameraAngleY, 0, 1, 0);

    // lines
    glColor3f(1.0f, 0.5f, 0.0f);
    glLineWidth(2.0f);
    glBegin(GL_LINES);

    int count = pipe.getPathCount();
    for(int i = 0; i < count-1; ++i)
    {
        glVertex3fv(&pipe.getPathPoint(i).x);
        glVertex3fv(&pipe.getPathPoint(i+1).x);
    }
    glEnd();

    // points
    glColor3f(0.0f, 1.0f, 1.0f);
    glPointSize(5.0f);
    glBegin(GL_POINTS);
    for(int i = 0; i < count; ++i)
    {
        glVertex3fv(&pipe.getPathPoint(i).x);
    }
    glEnd();
    glPointSize(1); // reset

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
}



///////////////////////////////////////////////////////////////////////////////
// draw 3D
///////////////////////////////////////////////////////////////////////////////
void draw()
{
    drawPath();
    drawPipe();
}



///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    initSharedMem();

    // register exit callback
    atexit(exitCB);

    // init GLUT and GL
    initGLUT(argc, argv);
    initGL();

    // check required extensions
    glExtension& ext = glExtension::getInstance();
    vboSupported = ext.isSupported("GL_ARB_vertex_buffer_object");
    if(vboSupported)
    {
        // create vertex buffer objects
        glGenBuffers(1, &vboId);
        glGenBuffers(1, &iboId);
        std::cout << "Video card supports GL_ARB_vertex_buffer_object." << std::endl;
    }
    else
    {
        std::cout << "[WARNING] Video card does NOT support GL_ARB_vertex_buffer_object." << std::endl;
    }

    glslSupported = ext.isSupported("GL_ARB_vertex_program") && ext.isSupported("GL_ARB_fragment_program");
    if(glslSupported)
    {
        std::cout << "Video card supports GLSL." << std::endl;
        // compile shaders and create GLSL program
        // If failed to create GLSL, reset flag to false
        glslSupported = initGLSL();
    }
    else
    {
        std::cout << "[WARNING] Video card does NOT support GLSL." << std::endl;
    }

    uniformLocationSupported = ext.isSupported("GL_ARB_explicit_uniform_location");
    if(uniformLocationSupported)
    {
        std::cout << "Video card supports GL_ARB_explicit_uniform_location." << std::endl;
    }
    else
    {
        std::cout << "[WARNING] Video card does NOT support GL_ARB_explicit_uniform_location." << std::endl;
    }

    // the last GLUT call (LOOP)
    // window will be shown and display callback is triggered by events
    // NOTE: this call never return main().
    glutMainLoop(); /* Start GLUT event-processing loop */

    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// initialize GLUT for windowing
///////////////////////////////////////////////////////////////////////////////
int initGLUT(int argc, char **argv)
{
    // GLUT stuff for windowing
    // initialization openGL window.
    // It must be called before any other GLUT routine.
    glutInit(&argc, argv);

    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_ALPHA);   // display mode

    glutInitWindowSize(screenWidth, screenHeight);              // window size

    glutInitWindowPosition(100, 100);                           // window location

    // finally, create a window with openGL context
    // Window will not displayed until glutMainLoop() is called
    // It returns a unique ID.
    int handle = glutCreateWindow(argv[0]);     // param is the title of window

    // register GLUT callback functions
    glutDisplayFunc(displayCB);
    glutTimerFunc(33, timerCB, 33);             // redraw only every given millisec
    //glutIdleFunc(idleCB);                       // redraw whenever system is idle
    glutReshapeFunc(reshapeCB);
    glutKeyboardFunc(keyboardCB);
    glutMouseFunc(mouseCB);
    glutMotionFunc(mouseMotionCB);
    glutPassiveMotionFunc(mousePassiveMotionCB);

    return handle;
}



///////////////////////////////////////////////////////////////////////////////
// initialize OpenGL
// disable unused features
///////////////////////////////////////////////////////////////////////////////
void initGL()
{
    glShadeModel(GL_SMOOTH);                    // shading mathod: GL_SMOOTH or GL_FLAT
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);      // 4-byte pixel alignment

    // enable /disable features
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_CULL_FACE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

     // track material ambient and diffuse from surface color, call it before glEnable(GL_COLOR_MATERIAL)
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    glClearColor(0, 0, 0, 0);                   // background color
    glClearStencil(0);                          // clear stencil buffer
    glClearDepth(1.0f);                         // 0 is near, 1 is far
    glDepthFunc(GL_LEQUAL);

    initLights();
    //setCamera(0, 0, 6, 0, 0, 0);
}



///////////////////////////////////////////////////////////////////////////////
// create glsl programs
///////////////////////////////////////////////////////////////////////////////
bool initGLSL()
{
    const int MAX_LENGTH = 2048;
    char log[MAX_LENGTH];
    int logLength = 0;
    int vsStatus, fsStatus;
    bool result;

    // create shader and program: flat shader ================================
    GLuint vsId1 = glCreateShader(GL_VERTEX_SHADER);
    GLuint fsId1 = glCreateShader(GL_FRAGMENT_SHADER);
    progId1 = glCreateProgram();

    // load shader sources
    glShaderSource(vsId1, 1, &vsSource1, NULL);
    glShaderSource(fsId1, 1, &fsSource1, NULL);

    // compile shader sources
    glCompileShader(vsId1);
    glCompileShader(fsId1);

    // attach shaders to the program
    glAttachShader(progId1, vsId1);
    glAttachShader(progId1, fsId1);

    // link program
    glLinkProgram(progId1);

    // get uniform/attrib locations
    glUseProgram(progId1);
    uniformMatrixModelViewProjection = glGetUniformLocation(progId1, "matrixModelViewProjection");
    attribVertexPosition = glGetAttribLocation(progId1, "vertexPosition");

    //@@ debug
    glGetShaderiv(vsId1, GL_COMPILE_STATUS, &vsStatus);
    if(vsStatus == GL_FALSE)
    {
        glGetShaderiv(vsId1, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(vsId1, MAX_LENGTH, &logLength, log);
        std::cout << "===== Vertex Shader Log #1 =====\n" << log << std::endl;
    }
    glGetShaderiv(fsId1, GL_COMPILE_STATUS, &fsStatus);
    if(fsStatus == GL_FALSE)
    {
        glGetShaderiv(fsId1, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(fsId1, MAX_LENGTH, &logLength, log);
        std::cout << "===== Fragment Shader Log #1 =====\n" << log << std::endl;
    }

    // create second shader and program: blinn + 2D texture ===================
    GLuint vsId2 = glCreateShader(GL_VERTEX_SHADER);
    GLuint fsId2 = glCreateShader(GL_FRAGMENT_SHADER);
    progId2 = glCreateProgram();
    glShaderSource(vsId2, 1, &vsSource2, NULL);
    glShaderSource(fsId2, 1, &fsSource2, NULL);
    glCompileShader(vsId2);
    glCompileShader(fsId2);

    // attach shaders to the program and link
    glAttachShader(progId2, vsId2);
    glAttachShader(progId2, fsId2);
    glLinkProgram(progId2);

    // get uniform/attrib locations
    glUseProgram(progId2);
    uniformMatrixModelView           = glGetUniformLocation(progId2, "matrixModelView");
    uniformMatrixModelViewProjection = glGetUniformLocation(progId2, "matrixModelViewProjection");
    uniformMatrixNormal              = glGetUniformLocation(progId2, "matrixNormal");
    uniformLightPosition             = glGetUniformLocation(progId2, "lightPosition");
    uniformLightAmbient              = glGetUniformLocation(progId2, "lightAmbient");
    uniformLightDiffuse              = glGetUniformLocation(progId2, "lightDiffuse");
    uniformLightSpecular             = glGetUniformLocation(progId2, "lightSpecular");
    uniformMaterialAmbient           = glGetUniformLocation(progId2, "materialAmbient");
    uniformMaterialDiffuse           = glGetUniformLocation(progId2, "materialDiffuse");
    uniformMaterialSpecular          = glGetUniformLocation(progId2, "materialSpecular");
    uniformMaterialShininess         = glGetUniformLocation(progId2, "materialShininess");
    uniformMap0                      = glGetUniformLocation(progId2, "map0");
    uniformTextureUsed               = glGetUniformLocation(progId2, "textureUsed");
    attribVertexPosition = glGetAttribLocation(progId2, "vertexPosition");
    attribVertexNormal   = glGetAttribLocation(progId2, "vertexNormal");
    attribVertexTexCoord = glGetAttribLocation(progId2, "vertexTexCoord");

    // set uniforms
    float lightAmbient[] = {0.3f, 0.3f, 0.3f, 1.0f};
    float lightDiffuse[] = {0.7f, 0.7f, 0.7f, 1.0f};
    float lightSpecular[] = {1.0f, 1.0f, 1.0f, 1.0f};
    float lightPosition[] = {0, 0, 1, 0};
    float materialAmbient[]  = {0.3f, 0.3f, 0.3f, 1};
    float materialDiffuse[]  = {1.0f, 1.0f, 0.0f, 1};
    float materialSpecular[] = {1.0f, 1.0f, 0.0f, 1};
    float materialShininess  = 16;
    glUniform4fv(uniformLightPosition, 1, lightPosition);
    glUniform4fv(uniformLightAmbient, 1, lightAmbient);
    glUniform4fv(uniformLightDiffuse, 1, lightDiffuse);
    glUniform4fv(uniformLightSpecular, 1, lightSpecular);
    glUniform4fv(uniformMaterialAmbient, 1, materialAmbient);
    glUniform4fv(uniformMaterialDiffuse, 1, materialDiffuse);
    glUniform4fv(uniformMaterialSpecular, 1, materialSpecular);
    glUniform1f(uniformMaterialShininess, materialShininess);
    glUniform1i(uniformMap0, 0);
    glUniform1i(uniformTextureUsed, 0); // false

    //@@ debug
    glGetShaderiv(vsId2, GL_COMPILE_STATUS, &vsStatus);
    if(vsStatus == GL_FALSE)
    {
        glGetShaderiv(vsId2, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(vsId2, MAX_LENGTH, &logLength, log);
        std::cout << "===== Vertex Shader Log #2 =====\n" << log << std::endl;
    }
    glGetShaderiv(fsId2, GL_COMPILE_STATUS, &fsStatus);
    if(fsStatus == GL_FALSE)
    {
        glGetShaderiv(fsId2, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(fsId2, MAX_LENGTH, &logLength, log);
        std::cout << "===== Fragment Shader Log #2 =====\n" << log << std::endl;
    }

    // unbind GLSL
    glUseProgram(0);

    // check GLSL status
    int linkStatus1, linkStatus2, linkStatus3;
    glGetProgramiv(progId1, GL_LINK_STATUS, &linkStatus1);
    glGetProgramiv(progId2, GL_LINK_STATUS, &linkStatus2);
    if(linkStatus1 == GL_FALSE)
    {
        glGetProgramiv(progId1, GL_INFO_LOG_LENGTH, &logLength);
        glGetProgramInfoLog(progId1, MAX_LENGTH, &logLength, log);
        std::cout << "===== GLSL Program Log #1 =====\n" << log << std::endl;
        result = false;
    }
    else if(linkStatus2 == GL_FALSE)
    {
        glGetProgramiv(progId2, GL_INFO_LOG_LENGTH, &logLength);
        glGetProgramInfoLog(progId2, MAX_LENGTH, &logLength, log);
        std::cout << "===== GLSL Program Log #2 =====\n" << log << std::endl;
        result = false;
    }
    else
    {
        result = true;
    }

    return result;
}



///////////////////////////////////////////////////////////////////////////////
// write 2d text using GLUT
// The projection matrix must be set to orthogonal before call this function.
///////////////////////////////////////////////////////////////////////////////
void drawString(const char *str, int x, int y, float color[4], void *font)
{
    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT); // lighting and color mask
    glDisable(GL_LIGHTING);     // need to disable lighting for proper text color
    glDisable(GL_TEXTURE_2D);

    glColor4fv(color);          // set text color
    glRasterPos2i(x, y);        // place text position

    // loop all characters in the string
    while(*str)
    {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glPopAttrib();
}



///////////////////////////////////////////////////////////////////////////////
// draw a string in 3D space
///////////////////////////////////////////////////////////////////////////////
void drawString3D(const char *str, float pos[3], float color[4], void *font)
{
    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT); // lighting and color mask
    glDisable(GL_LIGHTING);     // need to disable lighting for proper text color

    glColor4fv(color);          // set text color
    glRasterPos3fv(pos);        // place text position

    // loop all characters in the string
    while(*str)
    {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    glEnable(GL_LIGHTING);
    glPopAttrib();
}



///////////////////////////////////////////////////////////////////////////////
// initialize global variables
///////////////////////////////////////////////////////////////////////////////
bool initSharedMem()
{
    screenWidth = SCREEN_WIDTH;
    screenHeight = SCREEN_HEIGHT;
    mouseLeftDown = mouseRightDown = false;
    cameraAngleX = cameraAngleY = 0;
    drawMode = 0;

    // generate path for pipe
    path = buildSpiralPath(4, 1, -3, 3, PATH_TURNS, PATH_POINTS);
    std::cout << "fitst point: " << path[0] << std::endl;
    std::cout << "last point: " << path[path.size()-1] << std::endl;

    // sectional contour of pipe
    circle = buildCircle(0.5f, CIRCLE_SECTORS); // radius, segments

    // configure pipe
    std::vector<Vector3> p(1, path[0]);
    pipe.set(p, circle);
    currIndex = 0;

    return true;
}



///////////////////////////////////////////////////////////////////////////////
// clean up shared memory
///////////////////////////////////////////////////////////////////////////////
void clearSharedMem()
{
    // clean up VBOs
    if(vboSupported)
    {
        glDeleteBuffers(1, &vboId);
        glDeleteBuffers(1, &iboId);
        vboId = iboId = 0;
    }
}



///////////////////////////////////////////////////////////////////////////////
// initialize lights
///////////////////////////////////////////////////////////////////////////////
void initLights()
{
    // set up light colors (ambient, diffuse, specular)
    GLfloat lightKa[] = {0.0f, 0.0f, 0.0f, 1.0f};  // ambient light
    GLfloat lightKd[] = {1.0f, 1.0f, 1.0f, 1.0f};  // diffuse light
    GLfloat lightKs[] = {1, 1, 1, 1};           // specular light
    glLightfv(GL_LIGHT0, GL_AMBIENT, lightKa);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, lightKd);
    glLightfv(GL_LIGHT0, GL_SPECULAR, lightKs);

    // position the light
    float lightPos[4] = {0, 0, 1, 0}; // positional light
    glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

    glEnable(GL_LIGHT0);                        // MUST enable each light source after configuration
}



///////////////////////////////////////////////////////////////////////////////
// set camera position and lookat direction
///////////////////////////////////////////////////////////////////////////////
void setCamera(float posX, float posY, float posZ, float targetX, float targetY, float targetZ)
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(posX, posY, posZ, targetX, targetY, targetZ, 0, 1, 0); // eye(x,y,z), focal(x,y,z), up(x,y,z)
}



///////////////////////////////////////////////////////////////////////////////
// generate a spiral path along y-axis
// r1: starting radius
// r2: ending radius
// h1: starting height
// h2: ending height
// turns: # of revolutions
// points: # of points
std::vector<Vector3> buildSpiralPath(float r1, float r2, float h1, float h2,
                                     float turns, int points)
{
    const float PI = acos(-1.0f);
    std::vector<Vector3> vertices;
    Vector3 vertex;
    float r = r1;
    float rStep = (r2 - r1) / (points - 1);
    float y = h1;
    float yStep = (h2 - h1) / (points - 1);
    float a = 0;
    float aStep = (turns * 2 * PI) / (points - 1);
    for(int i = 0; i < points; ++i)
    {
        vertex.x = r * cos(a);
        vertex.z = r * sin(a);
        vertex.y = y;
        vertices.push_back(vertex);
        // next
        r += rStep;
        y += yStep;
        a += aStep;
    }
    return vertices;
}



///////////////////////////////////////////////////////////////////////////////
// generate points of a circle on xy plane
///////////////////////////////////////////////////////////////////////////////
std::vector<Vector3> buildCircle(float radius, int steps)
{
    std::vector<Vector3> points;
    if(steps < 2) return points;

    const float PI2 = acos(-1.0f) * 2.0f;
    float x, y, a;
    for(int i = 0; i <= steps; ++i)
    {
        a = PI2 / steps * i;
        x = radius * cosf(a);
        y = radius * sinf(a);
        points.push_back(Vector3(x, y, 0));
    }
    return points;
}



///////////////////////////////////////////////////////////////////////////////
// update vertex buffer data
///////////////////////////////////////////////////////////////////////////////
void updateVbo()
{
    if(!vboSupported) return;

    // compute buffer size and create tmp buffer
    std::vector< std::vector<Vector3> > contours = pipe.getContours();
    std::vector< std::vector<Vector3> > normals = pipe.getNormals();
    int contourCount = pipe.getContourCount();          // total # of contours
    int vertexCount = (int)pipe.getContour(0).size();   // # of vertices per contour
    int vertexSize = sizeof(float) * 3 * vertexCount;   // # of bytes per contour
    int offset = vertexSize * contourCount;             // offset where normals begin
    int bufferCount = 2 * 3 * vertexCount * contourCount; // total # of floats (vertex + normal)
    int bufferSize = sizeof(float) * bufferCount;       // total # of bytes
    std::vector<float> buffer(bufferCount);

    // CASE1: copy data to system memory first then copy to VBO at once
    if(contourCount == PATH_POINTS)
        t1.start();
    for(int i = 0; i < contourCount; ++i)
    {
        std::vector<Vector3> contour = contours[i]; // vertices per contour
        std::vector<Vector3> normal = normals[i];   // normals per contours
        int vIndex = i * 3 * vertexCount;
        int nIndex = (3 * vertexCount * contourCount) + vIndex;
        // copy # of bytes per contour
        std::memcpy(&buffer[vIndex], &contour[0].x, vertexSize);
        std::memcpy(&buffer[nIndex], &normal[0].x, vertexSize);
    }

    // copy to VBO using glBufferData()
    glBindBuffer(GL_ARRAY_BUFFER, vboId);
    glBufferData(GL_ARRAY_BUFFER, bufferSize, buffer.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    if(contourCount == PATH_POINTS)
    {
        t1.stop();
        std::cout << "BUFFER SIZE: " << bufferSize << " bytes" << std::endl;
        std::cout << "CASE 1: " << t1.getElapsedTimeInMilliSec() << " ms" << std::endl;
    }

    // CASE2: copy data to VBO directly using glBufferSubData() in loop
    if(contourCount == PATH_POINTS)
        t2.start();
    glBindBuffer(GL_ARRAY_BUFFER, vboId);
    glBufferData(GL_ARRAY_BUFFER, bufferSize, buffer.data(), GL_DYNAMIC_DRAW); // alloc
    for(int i = 0; i < contourCount; ++i)
    {
        std::vector<Vector3> contour = contours[i]; // vertices per contour
        std::vector<Vector3> normal = normals[i];   // normals per contours
        glBufferSubData(GL_ARRAY_BUFFER, (i * vertexSize), vertexSize, &contour[0].x);
        glBufferSubData(GL_ARRAY_BUFFER, offset + (i * vertexSize), vertexSize, &normal[0].x);
    }
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    if(contourCount == PATH_POINTS)
    {
        t2.stop();
        std::cout << "CASE 2: " << t2.getElapsedTimeInMilliSec() << " ms" << std::endl;
    }

    // build indices for triangle strip
    int k1 = 0, k2 = vertexCount;
    std::vector<unsigned int> indices;
    for(int i = 0; i < (contourCount-1) ; ++i)
    {
        for(int j = 0; j < vertexCount; ++j)
        {
            indices.push_back(k2++);
            indices.push_back(k1++);
        }
    }

    //std::cout << "INDEX SIZE: " << indices.size() <<std::endl;

    // copy indices to IBO
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned int), indices.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);


}



///////////////////////////////////////////////////////////////////////////////
// display info messages
///////////////////////////////////////////////////////////////////////////////
void showInfo()
{
    // backup current model-view matrix
    glPushMatrix();                     // save current modelview matrix
    glLoadIdentity();                   // reset modelview matrix

    // set to 2D orthogonal projection
    glMatrixMode(GL_PROJECTION);        // switch to projection matrix
    glPushMatrix();                     // save current projection matrix
    glLoadIdentity();                   // reset projection matrix
    gluOrtho2D(0, screenWidth, 0, screenHeight);  // set to orthogonal projection

    const int FONT_HEIGHT = 14;
    float color[4] = {1, 1, 1, 1};

    // for print infos
    std::stringstream ss;
    ss << std::fixed << std::setprecision(3);
    ss << "Press D to change drawing mode." << std::ends;
    drawString(ss.str().c_str(), 1, 1, color, font);
    ss.str("");

    // unset floating format
    ss << std::resetiosflags(std::ios_base::fixed | std::ios_base::floatfield);

    // restore projection matrix
    glPopMatrix();                   // restore to previous projection matrix

    // restore modelview matrix
    glMatrixMode(GL_MODELVIEW);      // switch to modelview matrix
    glPopMatrix();                   // restore to previous modelview matrix
}



///////////////////////////////////////////////////////////////////////////////
// set projection matrix as orthogonal
///////////////////////////////////////////////////////////////////////////////
void toOrtho()
{
    const float N = -1.0f;
    const float F = 1.0f;

    // set viewport to be the entire window
    glViewport(0, 0, (GLsizei)screenWidth, (GLsizei)screenHeight);

    // construct ortho projection matrix
    matrixProjection.identity();
    matrixProjection[0]  =  2 / screenWidth;
    matrixProjection[5]  =  2 / screenHeight;
    matrixProjection[10] = -2 / (F - N);
    matrixProjection[14] = -(F + N) / (F - N);

    // set orthographic viewing frustum
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(matrixProjection.get());
    //glLoadIdentity();
    //glOrtho(0, screenWidth, 0, screenHeight, -1, 1);

    // switch to modelview matrix in order to set scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}



///////////////////////////////////////////////////////////////////////////////
// set the projection matrix as perspective
///////////////////////////////////////////////////////////////////////////////
void toPerspective()
{
    const float N = 0.1f;
    const float F = 100.0f;
    const float DEG2RAD = 3.141592f / 180;
    const float FOV_Y = 40.0f * DEG2RAD;

    // set viewport to be the entire window
    glViewport(0, 0, (GLsizei)screenWidth, (GLsizei)screenHeight);

    // construct perspective projection matrix
    float aspectRatio = (float)(screenWidth) / screenHeight;
    float tangent = tanf(FOV_Y / 2.0f);     // tangent of half fovY
    float h = N * tangent;                  // half height of near plane
    float w = h * aspectRatio;              // half width of near plane
    matrixProjection.identity();
    matrixProjection[0]  =  N / w;
    matrixProjection[5]  =  N / h;
    matrixProjection[10] = -(F + N) / (F - N);
    matrixProjection[11] = -1;
    matrixProjection[14] = -(2 * F * N) / (F - N);
    matrixProjection[15] =  0;

    // set perspective viewing frustum
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(matrixProjection.get());
    //@@ equivalent fixed pipeline
    //glLoadIdentity();
    //gluPerspective(40.0f, (float)(screenWidth)/screenHeight, 0.1f, 100.0f); // FOV, AspectRatio, NearClip, FarClip

    // switch to modelview matrix in order to set scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}








//=============================================================================
// CALLBACKS
//=============================================================================

void displayCB()
{
    // add point
    if(animating)
    {
        ++currIndex;
        if(currIndex < path.size())
        {
            pipe.addPathPoint(path[currIndex]);
        }
        else
        {
            std::vector<Vector3> p(1, path[0]);
            pipe.set(p, circle);
            currIndex = 0;
        }

        // copy vertex data to VBO
        updateVbo();
    }

    // clear framebuffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    glPushMatrix();

    // draw 3D
    draw();

    showInfo();

    glPopMatrix();
    glutSwapBuffers();
}


void reshapeCB(int width, int height)
{
    screenWidth = width;
    screenHeight = height;
    toPerspective();
    std::cout << "window resized: " << width << " x " << height << std::endl;
}


void timerCB(int millisec)
{
    glutTimerFunc(millisec, timerCB, millisec);
    glutPostRedisplay();
}


void idleCB()
{
    glutPostRedisplay();
}


void keyboardCB(unsigned char key, int x, int y)
{
    switch(key)
    {
    case 27: // ESCAPE
        exit(0);
        break;

    case ' ':
        std::cout << "SPACE: " << animating << std::endl;
        animating = !animating;
        break;

    case 'd': // switch rendering modes (fill -> wire -> point)
    case 'D':
        drawMode = ++drawMode % 3;
        if(drawMode == 0)        // fill mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            glEnable(GL_DEPTH_TEST);
            glEnable(GL_CULL_FACE);
        }
        else if(drawMode == 1)  // wireframe mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
        }
        else                    // point mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
        }
        break;

    default:
        ;
    }

    glutPostRedisplay();
}


void mouseCB(int button, int state, int x, int y)
{
    mouseX = x;
    mouseY = y;

    if(button == GLUT_LEFT_BUTTON)
    {
        if(state == GLUT_DOWN)
        {
            mouseLeftDown = true;
        }
        else if(state == GLUT_UP)
        {
            mouseLeftDown = false;
        }
    }

    else if(button == GLUT_RIGHT_BUTTON)
    {
        if(state == GLUT_DOWN)
        {
            mouseRightDown = true;
        }
        else if(state == GLUT_UP)
        {
            mouseRightDown = false;
        }
    }
}


void mouseMotionCB(int x, int y)
{
    if(mouseLeftDown)
    {
        cameraAngleY += (x - mouseX);
        cameraAngleX += (y - mouseY);
        mouseX = x;
        mouseY = y;
    }
    if(mouseRightDown)
    {
        cameraDistance -= (y - mouseY) * 0.2f;
        mouseY = y;
    }
}


void mousePassiveMotionCB(int x, int y)
{
    mouseX = x;
    mouseY = y;
}



void exitCB()
{
    clearSharedMem();
}
